@@ -25,10 +25,13 @@ from typing import Tuple
 from typing import Type
 from typing import Optional
 from typing import Any
+from typing import Union
+from typing import TYPE_CHECKING
 
 import time
 import hashlib
 import os
+import re
 import socket
 import random
 import binascii
@@ -40,14 +43,18 @@ import libcloud.compute.ssh
 from libcloud.pricing import get_size_price
 from libcloud.compute.types import NodeState, StorageVolumeState,\
     DeploymentError
+if TYPE_CHECKING:
+    from libcloud.compute.deployment import Deployment
 from libcloud.compute.types import Provider
 from libcloud.compute.types import NodeImageMemberState
 from libcloud.compute.ssh import SSHClient
+from libcloud.compute.ssh import BaseSSHClient
 from libcloud.common.base import Connection
 from libcloud.common.base import ConnectionKey
 from libcloud.common.base import BaseDriver
 from libcloud.common.types import LibcloudError
 from libcloud.compute.ssh import have_paramiko
+from libcloud.compute.ssh import SSHCommandTimeoutError
 
 from libcloud.utils.networking import is_private_subnet
 from libcloud.utils.networking import is_valid_ip_address
@@ -62,6 +69,10 @@ else:
     SSH_TIMEOUT_EXCEPTION_CLASSES = (IOError, socket.gaierror,  # type: ignore
                                      socket.error)  # type: ignore
 
+T_Auth = Union['NodeAuthSSHKey', 'NodeAuthPassword']
+
+T_Ssh_key = Union[List[str], str]
+
 # How long to wait for the node to come online after creating it
 NODE_ONLINE_WAIT_TIMEOUT = 10 * 60
 
@@ -69,6 +80,22 @@ NODE_ONLINE_WAIT_TIMEOUT = 10 * 60
 # script.
 SSH_CONNECT_TIMEOUT = 5 * 60
 
+# Error message which should be considered fatal for deploy_node() method and
+# on which we should abort retrying and immediately propagate the error
+SSH_FATAL_ERROR_MSGS = [
+    # Propagate (key) file doesn't exist errors
+    # NOTE: Paramiko only supports PEM private key format
+    # See https://github.com/paramiko/paramiko/issues/1313
+    # for details
+    'no such file or directory',
+    'invalid key',
+    'not a valid ',
+    'invalid or unsupported key type',
+    'private file is encrypted',
+    'private key file is encrypted',
+    'private key file checkints do not match',
+    'invalid password provided'
+]
 
 __all__ = [
     'Node',
@@ -238,6 +265,7 @@ class Node(UuidMixin):
         UuidMixin.__init__(self)
 
     def reboot(self):
+        # type: () -> bool
         """
         Reboot this node
 
@@ -278,6 +306,7 @@ class Node(UuidMixin):
         return self.driver.stop_node(self)
 
     def destroy(self):
+        # type: () -> bool
         """
         Destroy this node
 
@@ -339,7 +368,7 @@ class NodeSize(UuidMixin):
                  disk,  # type: int
                  bandwidth,  # type: Optional[int]
                  price,  # type: float
-                 driver,  # type: Type[NodeDriver]
+                 driver,  # type: NodeDriver
                  extra=None  # type: Optional[dict]
                  ):
         """
@@ -410,7 +439,7 @@ class NodeImage(UuidMixin):
     def __init__(self,
                  id,  # type: str
                  name,  # type: str
-                 driver,  # type: Type[NodeDriver]
+                 driver,  # type: NodeDriver
                  extra=None  # type: Optional[dict]
                  ):
         """
@@ -456,7 +485,7 @@ class NodeImageMember(UuidMixin):
                  id,  # type: str
                  image_id,  # type: str
                  state,  # type: NodeImageMemberState
-                 driver,  # type: Type[NodeDriver]
+                 driver,  # type: NodeDriver
                  created=None,  # type: datetime.datetime
                  extra=None  # type: Optional[dict]
                  ):
@@ -511,7 +540,7 @@ class NodeLocation(object):
                  id,  # type: str
                  name,  # type: str
                  country,  # type: str
-                 driver,  # type: Type[NodeDriver]
+                 driver,  # type: NodeDriver
                  extra=None  # type: Optional[dict]
                  ):
         """
@@ -559,7 +588,7 @@ class NodeAuthSSHKey(object):
     def __init__(self, pubkey):
         # type: (str) -> None
         """
-        :param pubkey: Public key matetiral.
+        :param pubkey: Public key material.
         :type pubkey: ``str``
         """
         self.pubkey = pubkey
@@ -573,6 +602,7 @@ class NodeAuthPassword(object):
     A password to be used for authentication to a node.
     """
     def __init__(self, password, generated=False):
+        # type: (str, bool) -> None
         """
         :param password: Password.
         :type password: ``str``
@@ -592,8 +622,15 @@ class StorageVolume(UuidMixin):
     A base StorageVolume class to derive from.
     """
 
-    def __init__(self, id, name, size, driver,
-                 state=None, extra=None):
+    def __init__(self,
+                 id,  # type: str
+                 name,  # type: str
+                 size,  # type: int
+                 driver,  # type: NodeDriver
+                 state=None,  # type: Optional[StorageVolumeState]
+                 extra=None  # type: Optional[Dict]
+                 ):
+        # type: (...) -> None
         """
         :param id: Storage volume ID.
         :type id: ``str``
@@ -623,12 +660,14 @@ class StorageVolume(UuidMixin):
         UuidMixin.__init__(self)
 
     def list_snapshots(self):
+        # type: () -> List[VolumeSnapshot]
         """
         :rtype: ``list`` of ``VolumeSnapshot``
         """
         return self.driver.list_volume_snapshots(volume=self)
 
     def attach(self, node, device=None):
+        # type: (Node, Optional[str]) -> bool
         """
         Attach this volume to a node.
 
@@ -642,20 +681,20 @@ class StorageVolume(UuidMixin):
         :return: ``True`` if attach was successful, ``False`` otherwise.
         :rtype: ``bool``
         """
-
         return self.driver.attach_volume(node=node, volume=self, device=device)
 
     def detach(self):
+        # type: () -> bool
         """
         Detach this volume from its node
 
         :return: ``True`` if detach was successful, ``False`` otherwise.
         :rtype: ``bool``
         """
-
         return self.driver.detach_volume(volume=self)
 
     def snapshot(self, name):
+        # type: (str) -> VolumeSnapshot
         """
         Creates a snapshot of this volume.
 
@@ -665,6 +704,7 @@ class StorageVolume(UuidMixin):
         return self.driver.create_volume_snapshot(volume=self, name=name)
 
     def destroy(self):
+        # type: () -> bool
         """
         Destroy this storage volume.
 
@@ -683,8 +723,16 @@ class VolumeSnapshot(object):
     """
     A base VolumeSnapshot class to derive from.
     """
-    def __init__(self, id, driver, size=None, extra=None, created=None,
-                 state=None, name=None):
+    def __init__(self,
+                 id,  # type: str
+                 driver,  # type: NodeDriver
+                 size=None,  # type: int
+                 extra=None,  # type: Optional[Dict]
+                 created=None,  # type: Optional[datetime.datetime]
+                 state=None,  # type: StorageVolumeState
+                 name=None  # type: Optional[str]
+                 ):
+        # type: (...) -> None
         """
         VolumeSnapshot constructor.
 
@@ -707,7 +755,7 @@ class VolumeSnapshot(object):
 
         :param      state: A string representing the state the snapshot is
                            in. See `libcloud.compute.types.StorageVolumeState`.
-        :type       state: ``str``
+        :type       state: ``StorageVolumeState``
 
         :param      name: A string representing the name of the snapshot
         :type       name: ``str``
@@ -721,6 +769,7 @@ class VolumeSnapshot(object):
         self.name = name
 
     def destroy(self):
+        # type: () -> bool
         """
         Destroys this snapshot.
 
@@ -738,8 +787,15 @@ class KeyPair(object):
     Represents a SSH key pair.
     """
 
-    def __init__(self, name, public_key, fingerprint, driver, private_key=None,
-                 extra=None):
+    def __init__(self,
+                 name,  # type: str
+                 public_key,  # type: str
+                 fingerprint,  # type: str
+                 driver,  # type: NodeDriver
+                 private_key=None,  # type: Optional[str]
+                 extra=None  # type: Optional[Dict]
+                 ):
+        # type: (...) -> None
         """
         Constructor.
 
@@ -784,7 +840,7 @@ class NodeDriver(BaseDriver):
     name = None  # type: str
     api_name = None  # type: str
     website = None  # type: str
-    type = None  # type: Provider
+    type = None  # type: Union[Provider,str]
     port = None  # type: int
     features = {'create_node': []}  # type: Dict[str, List[str]]
 
@@ -838,7 +894,13 @@ class NodeDriver(BaseDriver):
         raise NotImplementedError(
             'list_locations not implemented for this driver')
 
-    def create_node(self, **kwargs):
+    def create_node(self,
+                    name,  # type: str
+                    size,  # type: NodeSize
+                    image,  # type: NodeImage
+                    location=None,  # type: Optional[NodeLocation]
+                    auth=None  # type: T_Auth
+                    ):
         # type: (...) -> Node
         """
         Create a new node instance. This instance will be started
@@ -922,7 +984,19 @@ class NodeDriver(BaseDriver):
         raise NotImplementedError(
             'create_node not implemented for this driver')
 
-    def deploy_node(self, **kwargs):
+    def deploy_node(self,
+                    deploy,  # type: Deployment
+                    ssh_username='root',  # type: str
+                    ssh_alternate_usernames=None,  # type: Optional[List[str]]
+                    ssh_port=22,  # type: int
+                    ssh_timeout=10,  # type: int
+                    ssh_key=None,  # type: Optional[T_Ssh_key]
+                    ssh_key_password=None,  # type: Optional[str]
+                    auth=None,  # type: T_Auth
+                    timeout=SSH_CONNECT_TIMEOUT,  # type: int
+                    max_tries=3,  # type: int
+                    ssh_interface='public_ips',  # type: str
+                    **create_node_kwargs):
         # type: (...) -> Node
         """
         Create a new node, and start deployment.
@@ -997,6 +1071,9 @@ class NodeDriver(BaseDriver):
                              to attempt to authenticate. (optional)
         :type ssh_key: ``str`` or ``list`` of ``str``
 
+        :param ssh_key_password: Optional password used for encrypted keys.
+        :type ssh_key_password: ``str``
+
         :param timeout: How many seconds to wait before timing out.
                              (default is 600)
         :type timeout: ``int``
@@ -1013,13 +1090,12 @@ class NodeDriver(BaseDriver):
             raise RuntimeError('paramiko is not installed. You can install ' +
                                'it using pip: pip install paramiko')
 
-        if 'auth' in kwargs:
-            auth = kwargs['auth']
+        if auth:
             if not isinstance(auth, (NodeAuthSSHKey, NodeAuthPassword)):
                 raise NotImplementedError(
                     'If providing auth, only NodeAuthSSHKey or'
                     'NodeAuthPassword is supported')
-        elif 'ssh_key' in kwargs:
+        elif ssh_key:
             # If an ssh_key is provided we can try deploy_node
             pass
         elif 'create_node' in self.features:
@@ -1031,45 +1107,78 @@ class NodeDriver(BaseDriver):
             raise NotImplementedError(
                 'deploy_node not implemented for this driver')
 
-        node = self.create_node(**kwargs)
-        max_tries = kwargs.get('max_tries', 3)
+        # NOTE 1: This is a workaround for legacy code. Sadly a lot of legacy
+        # code uses **kwargs in "create_node()" method and simply ignores
+        # "deploy_node()" arguments which are passed to it.
+        # That's obviously far from idea that's why we first try to pass only
+        # non-deploy node arguments to the "create_node()" methods and if it
+        # that doesn't work, fall back to the old approach and simply pass in
+        # all the arguments
+        # NOTE 2: Some drivers which use password based SSH authentication
+        # rely on password being stored on the "auth" argument and that's why
+        # we also propagate that argument to "create_node()" method.
+        try:
+            # NOTE: We only pass auth to the method if auth argument is
+            # provided
+            if auth:
+                node = self.create_node(auth=auth, **create_node_kwargs)
+            else:
+                node = self.create_node(**create_node_kwargs)
+        except TypeError as e:
+            msg_1_re = (r'create_node\(\) missing \d+ required '
+                        'positional arguments.*')
+            msg_2_re = r'create_node\(\) takes at least \d+ arguments.*'
+            if re.match(msg_1_re, str(e)) or re.match(msg_2_re, str(e)):
+                # pylint: disable=unexpected-keyword-arg
+                node = self.create_node(  # type: ignore
+                    deploy=deploy,
+                    ssh_username=ssh_username,
+                    ssh_alternate_usernames=ssh_alternate_usernames,
+                    ssh_port=ssh_port,
+                    ssh_timeout=ssh_timeout,
+                    ssh_key=ssh_key,
+                    auth=auth,
+                    timeout=timeout,
+                    max_tries=max_tries,
+                    ssh_interface=ssh_interface,
+                    **create_node_kwargs)
+                # pylint: enable=unexpected-keyword-arg
+            else:
+                raise e
 
         password = None
-        if 'auth' in kwargs:
-            if isinstance(kwargs['auth'], NodeAuthPassword):
-                password = kwargs['auth'].password
+        if auth:
+            if isinstance(auth, NodeAuthPassword):
+                password = auth.password
         elif 'password' in node.extra:
             password = node.extra['password']
 
-        ssh_interface = kwargs.get('ssh_interface', 'public_ips')
+        wait_timeout = timeout or NODE_ONLINE_WAIT_TIMEOUT
 
         # Wait until node is up and running and has IP assigned
         try:
             node, ip_addresses = self.wait_until_running(
                 nodes=[node],
                 wait_period=3,
-                timeout=kwargs.get('timeout', NODE_ONLINE_WAIT_TIMEOUT),
+                timeout=wait_timeout,
                 ssh_interface=ssh_interface)[0]
         except Exception as e:
             raise DeploymentError(node=node, original_exception=e, driver=self)
 
-        ssh_username = kwargs.get('ssh_username', 'root')
-        ssh_alternate_usernames = kwargs.get('ssh_alternate_usernames', [])
-        ssh_port = kwargs.get('ssh_port', 22)
-        ssh_timeout = kwargs.get('ssh_timeout', 10)
-        ssh_key_file = kwargs.get('ssh_key', None)
-        timeout = kwargs.get('timeout', SSH_CONNECT_TIMEOUT)
+        ssh_alternate_usernames = ssh_alternate_usernames or []
+        deploy_timeout = timeout or SSH_CONNECT_TIMEOUT
 
         deploy_error = None
 
         for username in ([ssh_username] + ssh_alternate_usernames):
             try:
                 self._connect_and_run_deployment_script(
-                    task=kwargs['deploy'], node=node,
+                    task=deploy, node=node,
                     ssh_hostname=ip_addresses[0], ssh_port=ssh_port,
                     ssh_username=username, ssh_password=password,
-                    ssh_key_file=ssh_key_file, ssh_timeout=ssh_timeout,
-                    timeout=timeout, max_tries=max_tries)
+                    ssh_key_file=ssh_key, ssh_key_password=ssh_key_password,
+                    ssh_timeout=ssh_timeout,
+                    timeout=deploy_timeout, max_tries=max_tries)
             except Exception as e:
                 # Try alternate username
                 # Todo: Need to fix paramiko so we can catch a more specific
@@ -1455,9 +1564,14 @@ class NodeDriver(BaseDriver):
         raise NotImplementedError(
             'delete_key_pair not implemented for this driver')
 
-    def wait_until_running(self, nodes, wait_period=3,
-                           timeout=600, ssh_interface='public_ips',
-                           force_ipv4=True, ex_list_nodes_kwargs=None):
+    def wait_until_running(self,
+                           nodes,  # type: List[Node]
+                           wait_period=3,  # type: float
+                           timeout=600,  # type: int
+                           ssh_interface='public_ips',  # type: str
+                           force_ipv4=True,  # type: bool
+                           ex_list_nodes_kwargs=None  # type: Optional[Dict]
+                           ):
         # type: (...) -> List[Tuple[Node, List[str]]]
         """
         Block until the provided nodes are considered running.
@@ -1513,7 +1627,7 @@ class NodeDriver(BaseDriver):
             return [address for address in addresses if is_supported(address)]
 
         if ssh_interface not in ['public_ips', 'private_ips']:
-            raise ValueError('ssh_interface argument must either be' +
+            raise ValueError('ssh_interface argument must either be ' +
                              'public_ips or private_ips')
 
         start = time.time()
@@ -1551,6 +1665,7 @@ class NodeDriver(BaseDriver):
                             driver=self)
 
     def _get_and_check_auth(self, auth):
+        # type: (T_Auth) -> T_Auth
         """
         Helper function for providers supporting :class:`.NodeAuthPassword` or
         :class:`.NodeAuthSSHKey`
@@ -1599,6 +1714,7 @@ class NodeDriver(BaseDriver):
 
     def _wait_until_running(self, node, wait_period=3, timeout=600,
                             ssh_interface='public_ips', force_ipv4=True):
+        # type: (Node, float, int, str, bool) -> List[Tuple[Node, List[str]]]
         # This is here for backward compatibility and will be removed in the
         # next major release
         return self.wait_until_running(nodes=[node], wait_period=wait_period,
@@ -1607,6 +1723,7 @@ class NodeDriver(BaseDriver):
                                        force_ipv4=force_ipv4)
 
     def _ssh_client_connect(self, ssh_client, wait_period=1.5, timeout=300):
+        # type: (BaseSSHClient, float, int) -> BaseSSHClient
         """
         Try to connect to the remote SSH server. If a connection times out or
         is refused it is retried up to timeout number of seconds.
@@ -1632,20 +1749,11 @@ class NodeDriver(BaseDriver):
                 ssh_client.connect()
             except SSH_TIMEOUT_EXCEPTION_CLASSES as e:
                 # Errors which represent fatal invalid key files which should
-                # be propagated to the user
+                # be propagated to the user without us retrying
                 message = str(e).lower()
-                invalid_key_msgs = [
-                    'no such file or directory',
-                    'invalid key',
-                    'not a valid ',
-                ]
 
-                # Propagate (key) file doesn't exist errors
-                # NOTE: Paramiko only supports PEM private key format
-                # See https://github.com/paramiko/paramiko/issues/1313
-                # for details
-                for invalid_key_msg in invalid_key_msgs:
-                    if invalid_key_msg in message:
+                for fatal_msg in SSH_FATAL_ERROR_MSGS:
+                    if fatal_msg in message:
                         raise e
 
                 # Retry if a connection is refused, timeout occurred,
@@ -1659,10 +1767,20 @@ class NodeDriver(BaseDriver):
         raise LibcloudError(value='Could not connect to the remote SSH ' +
                             'server. Giving up.', driver=self)
 
-    def _connect_and_run_deployment_script(self, task, node, ssh_hostname,
-                                           ssh_port, ssh_username,
-                                           ssh_password, ssh_key_file,
-                                           ssh_timeout, timeout, max_tries):
+    def _connect_and_run_deployment_script(
+        self,
+        task,  # type: Deployment
+        node,  # type: Node
+        ssh_hostname,  # type: str
+        ssh_port,  # type: int
+        ssh_username,  # type: str
+        ssh_password,  # type: Optional[str]
+        ssh_key_file,  # type: Optional[T_Ssh_key]
+        ssh_key_password,  # type: Optional[str]
+        ssh_timeout,  # type: int
+        timeout,  # type: int
+        max_tries  # type: int
+    ):
         """
         Establish an SSH connection to the node and run the provided deployment
         task.
@@ -1672,7 +1790,7 @@ class NodeDriver(BaseDriver):
         """
         ssh_client = SSHClient(hostname=ssh_hostname,
                                port=ssh_port, username=ssh_username,
-                               password=ssh_password,
+                               password=ssh_key_password or ssh_password,
                                key_files=ssh_key_file,
                                timeout=ssh_timeout)
 
@@ -1686,6 +1804,7 @@ class NodeDriver(BaseDriver):
         return node
 
     def _run_deployment_script(self, task, node, ssh_client, max_tries=3):
+        # type: (Deployment, Node, BaseSSHClient, int) -> Node
         """
         Run the deployment script on the provided node. At this point it is
         assumed that SSH connection has already been established.
@@ -1711,6 +1830,9 @@ class NodeDriver(BaseDriver):
         while tries < max_tries:
             try:
                 node = task.run(node, ssh_client)
+            except SSHCommandTimeoutError as e:
+                # Command timeout exception is fatal so we don't retry it.
+                raise e
             except Exception as e:
                 tries += 1
 
@@ -1722,7 +1844,10 @@ class NodeDriver(BaseDriver):
                 ssh_client.close()
                 return node
 
+        return node
+
     def _get_size_price(self, size_id):
+        # type: (str) -> float
         """
         Return pricing information for the provided size id.
         """
