@@ -335,27 +335,6 @@ class S3MockHttp(MockHttp):
                 headers,
                 httplib.responses[httplib.OK])
 
-    def _foo_bar_container_foo_test_upload_INVALID_HASH1(self, method, url,
-                                                         body, headers):
-        body = ''
-        headers = {}
-        headers['etag'] = '"foobar"'
-        # test_upload_object_invalid_hash1
-        return (httplib.OK,
-                body,
-                headers,
-                httplib.responses[httplib.OK])
-
-    def _foo_bar_container_foo_test_upload_INVALID_HASH2(self, method, url,
-                                                         body, headers):
-        # test_upload_object_invalid_hash2
-        body = ''
-        headers = {'etag': '"hash343hhash89h932439jsaa89"'}
-        return (httplib.OK,
-                body,
-                headers,
-                httplib.responses[httplib.OK])
-
     def _foo_bar_container_foo_test_upload(self, method, url, body, headers):
         # test_upload_object_success
         body = ''
@@ -772,12 +751,11 @@ class S3Tests(unittest.TestCase):
         def upload_file(self, object_name=None, content_type=None,
                         request_path=None, request_method=None,
                         headers=None, file_path=None, stream=None):
-            return {'response': make_response(200),
+            headers = {'etag': '"foobar"'}
+            return {'response': make_response(200, headers=headers),
                     'bytes_transferred': 1000,
                     'data_hash': 'hash343hhash89h932439jsaa89'}
 
-        self.mock_response_klass.type = 'INVALID_HASH1'
-
         old_func = self.driver_type._upload_object
         self.driver_type._upload_object = upload_file
         file_path = os.path.abspath(__file__)
@@ -802,12 +780,11 @@ class S3Tests(unittest.TestCase):
         def upload_file(self, object_name=None, content_type=None,
                         request_path=None, request_method=None,
                         headers=None, file_path=None, stream=None):
-            return {'response': make_response(200, headers={'etag': 'woopwoopwoop'}),
+            headers = {'etag': '"hash343hhash89h932439jsaa89"'}
+            return {'response': make_response(200, headers=headers),
                     'bytes_transferred': 1000,
                     'data_hash': '0cc175b9c0f1b6a831c399e269772661'}
 
-        self.mock_response_klass.type = 'INVALID_HASH2'
-
         old_func = self.driver_type._upload_object
         self.driver_type._upload_object = upload_file
 
@@ -827,6 +804,31 @@ class S3Tests(unittest.TestCase):
         finally:
             self.driver_type._upload_object = old_func
 
+    def test_upload_object_invalid_hash_kms_encryption(self):
+        # Hash check should be skipped when AWS KMS server side encryption is
+        # used
+        def upload_file(self, object_name=None, content_type=None,
+                        request_path=None, request_method=None,
+                        headers=None, file_path=None, stream=None):
+            headers = {'etag': 'blahblah', 'x-amz-server-side-encryption': 'aws:kms'}
+            return {'response': make_response(200, headers=headers),
+                    'bytes_transferred': 1000,
+                    'data_hash': 'hash343hhash89h932439jsaa81'}
+
+        old_func = self.driver_type._upload_object
+        self.driver_type._upload_object = upload_file
+        file_path = os.path.abspath(__file__)
+        container = Container(name='foo_bar_container', extra={},
+                              driver=self.driver)
+        object_name = 'foo_test_upload'
+        try:
+            self.driver.upload_object(file_path=file_path, container=container,
+                                      object_name=object_name,
+                                      verify_hash=True)
+        finally:
+            self.driver_type._upload_object = old_func
+
+
     def test_upload_object_success(self):
         def upload_file(self, object_name=None, content_type=None,
                         request_path=None, request_method=None,
@@ -1071,6 +1073,14 @@ class S3Tests(unittest.TestCase):
         self.assertRaisesRegexp(ValueError, expected_msg, S3StorageDriver,
                                 *self.driver_args, region='foo')
 
+        # host argument still has precedence over reguin
+        driver3  = S3StorageDriver(*self.driver_args, region='ap-south-1', host='host1.bar.com')
+        self.assertEqual(driver3.region, 'ap-south-1')
+        self.assertEqual(driver3.connection.host, 'host1.bar.com')
+
+        driver4  = S3StorageDriver(*self.driver_args, host='host2.bar.com')
+        self.assertEqual(driver4.connection.host, 'host2.bar.com')
+
     def test_deprecated_driver_class_per_region(self):
         driver = S3USWestStorageDriver(*self.driver_args)
         self.assertEqual(driver.region, 'us-west-1')
